﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Korero.Communication;
using Korero.Properties;
using Korero.Serialization;
using Korero.Utilities;
using Serilog;

namespace Korero.Selection
{
    public partial class GroupSelectionForm : Form
    {
        #region Public Events & Delegates

        public event EventHandler<GroupSelectedEventArgs> GroupSelected;

        #endregion

        #region Private Delegates, Events, Enums, Properties, Indexers and Fields

        private readonly CancellationToken _cancellationToken;

        private readonly CancellationTokenSource _cancellationTokenSource;

        private readonly dynamic _data;

        private readonly MqttCommunication _mqttCommunication;

        #endregion

        #region Constructors, Destructors and Finalizers

        public GroupSelectionForm()
        {
            InitializeComponent();

            _cancellationTokenSource = new CancellationTokenSource();
            _cancellationToken = _cancellationTokenSource.Token;
        }

        public GroupSelectionForm(MqttCommunication mqttCommunication, dynamic data) : this()
        {
            _data = data;
            _mqttCommunication = mqttCommunication;
        }

        public GroupSelectionForm(MqttCommunication mqttCommunication) : this()
        {
            _mqttCommunication = mqttCommunication;
        }

        /// <summary>
        ///     Clean up any resources being used.
        /// </summary>
        /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
        protected override void Dispose(bool disposing)
        {
            if (disposing && components != null)
            {
                components.Dispose();
            }

            base.Dispose(disposing);
        }

        #endregion

        #region Event Handlers
        private void GroupSelectionForm_Load(object sender, EventArgs e)
        {
            Utilities.WindowState.FormTracker.Track(this);
        }
        private async void GroupSelectionForm_Shown(object sender, EventArgs e)
        {
            await RetrieveCurrentGroups();
        }

        private void GroupsListBox_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            EmitSelection((ListBox) sender, e);
        }

        private void GroupSelectionForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            _cancellationTokenSource.Cancel();
        }

        private async void Button1_Click(object sender, EventArgs e)
        {
            var name = textBox1.Text;

            var data = new Dictionary<string, string>
            {
                {"command", "batchgroupnametokey"},
                {"group", Settings.Default.Group},
                {"password", Settings.Default.Password},
                {"groups", new CSV(new[] {$"{name}"})},
                {"sift", new CSV(new[] {"take", "1"})}
            };

            var callback = await _mqttCommunication.SendCommand(new Command(data), _cancellationToken);

            if (callback == null || !callback.Success)
            {
                if (callback != null)
                {
                    Log.Warning("Command {Command} has failed with {Error}.",
                        callback.Command, callback.Error);
                }

                return;
            }

            if (!callback.Contains($"{name}"))
            {
                return;
            }

            var groupSelection = new GroupSelection(name);

            GroupSelected?.Invoke(this, new GroupSelectedEventArgs(groupSelection, _data));

            Close();
        }

        #endregion

        #region Public Methods

        public void EmitSelection(ListBox listBox, MouseEventArgs e)
        {
            var index = listBox.IndexFromPoint(e.Location);
            if (index == ListBox.NoMatches)
            {
                return;
            }

            var groupSelection = (GroupSelection) listBox.Items[index];
            GroupSelected?.Invoke(this, new GroupSelectedEventArgs(groupSelection, _data));

            Close();
        }

        #endregion

        #region Private Methods

        private async Task RetrieveCurrentGroups()
        {
            var data = new Dictionary<string, string>
            {
                {"command", "getcurrentgroups"},
                {"group", Settings.Default.Group},
                {"password", Settings.Default.Password},
                {"sift", new CSV(new[] {"each", "2"})}
            };

            var callback = await _mqttCommunication.SendCommand(new Command(data), _cancellationToken);

            if (callback == null || !callback.Success)
            {
                if (callback != null)
                {
                    Log.Warning("Command {Command} has failed with {Error}.",
                        callback.Command, callback.Error);
                }

                return;
            }

            foreach (var item in callback.Data)
            {
                groupsListBox.InvokeIfRequired(listBox => { listBox.Items.Add(new GroupSelection(item)); });
            }
        }

        #endregion


    }
}