﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Threading.Tasks;
using System.Windows.Forms;
using Microsoft.Win32;

namespace Korero.Utilities
{
    public static class Miscellaneous
    {
        #region Public Methods

        public static TimeSpan GetIdleTime()
        {
            var lastInPut = new Natives.LASTINPUTINFO();
            lastInPut.cbSize = (uint) Marshal.SizeOf(lastInPut);
            Natives.GetLastInputInfo(ref lastInPut);

            return TimeSpan.FromMilliseconds((uint) Environment.TickCount - lastInPut.dwTime);
        }

        public static bool LaunchOnBootSet(bool enable)
        {
            using (var key = Registry.CurrentUser.OpenSubKey
                ("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run", true))
            {
                if (key == null)
                {
                    return false;
                }

                switch (enable)
                {
                    case true:
                        key.SetValue(Constants.AssemblyName, Assembly.GetEntryAssembly().Location);
                        break;
                    default:
                        key.DeleteValue(Constants.AssemblyName, false);
                        break;
                }
            }

            return true;
        }

        public static bool LaunchOnBootGet()
        {
            using (var key = Registry.CurrentUser.OpenSubKey
                ("SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run", true))
            {
                return key?.GetValue(Constants.AssemblyName) != null;
            }
        }

        /// <summary>
        ///     Enable double buffering for an arbitrary control.
        /// </summary>
        /// <param name="control">the control to enable double buffering for</param>
        /// <returns>true on success</returns>
        /// <remarks>Do not enable double buffering on RDP: https://devblogs.microsoft.com/oldnewthing/20060103-12/?p=32793</remarks>
        public static bool SetDoubleBuffered(this Control control)
        {
            if (SystemInformation.TerminalServerSession)
            {
                return false;
            }

            var dgvType = control.GetType();
            var pi = dgvType.GetProperty("DoubleBuffered",
                BindingFlags.Instance | BindingFlags.NonPublic);
            if (pi == null)
            {
                return false;
            }

            pi.SetValue(control, true, null);

            return true;
        }

        public static async Task<Icon> CreateIconFromResource(string resource)
        {
            var iconBytes = await LoadResource(resource);
            using (var iconMemoryStream = new MemoryStream(iconBytes))
            {
                var bitmap = (Bitmap) Image.FromStream(iconMemoryStream);
                var bitmapIntPtr = bitmap.GetHicon();
                var icon = Icon.FromHandle(bitmapIntPtr);
                return icon;
            }
        }

        public static async Task<byte[]> LoadResource(string resource)
        {
            var assembly = Assembly.GetExecutingAssembly();

            using (var manifestResourceStream = assembly.GetManifestResourceStream(resource))
            {
                if (manifestResourceStream == null)
                {
                    return null;
                }

                var memoryStream = new MemoryStream();

                await manifestResourceStream.CopyToAsync(memoryStream);

                memoryStream.Position = 0L;

                return memoryStream.ToArray();
            }
        }

        public static void InvokeIfRequired<T>(this T control, Action<T> action) where T : Control
        {
            if (control.InvokeRequired)
            {
                control.BeginInvoke((MethodInvoker) delegate { action(control); });
                return;
            }

            action(control);
        }

        public static T MapValueToRange<T>(this T value, T xMin, T xMax, T yMin, T yMax)
            where T : struct, IComparable<T>, IConvertible
        {
            return (dynamic) yMin +
                   ((dynamic) yMax - (dynamic) yMin) * ((dynamic) value - (dynamic) xMin) /
                   ((dynamic) xMax - (dynamic) xMin);
        }

        public static IEnumerable<TU> SequenceSubtract<TU, TV>(this IEnumerable<TU> a,
                                                               IEnumerable<TV> b,
                                                               Func<TU, TV, bool> cmp)
        {
            var eb = new List<TV>(b);

            using (var ea = a.GetEnumerator())
            {
                while (ea.MoveNext())
                {
                    if (ea.Current == null)
                    {
                        continue;
                    }

                    foreach (var ib in eb)
                    {
                        if (cmp.Invoke(ea.Current, ib))
                        {
                            continue;
                        }

                        yield return ea.Current;

                        break;
                    }
                }
            }
        }

        #endregion
    }
}